import json
from abc import ABC, abstractmethod
from functools import wraps
from logging import getLogger
from typing import Tuple, Dict, Optional, Union, List, Callable, Any

from surrealist.errors import OperationOnClosedConnectionError, TooManyNestedLevelsError
from surrealist.result import SurrealResult
from surrealist.utils import DEFAULT_TIMEOUT, crop_data

logger = getLogger("connection")
LINK = "https://github.com/kotolex/surrealist?tab=readme-ov-file#recursion-and-json-in-python"


def connected(func):
    """
    Decorator for methods to make sure the underlying connection is alive (connected to DB)

    :param func: method to decorate
    :raise OperationOnClosedConnectionError: if connection is already closed
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        # args[0] is a self-argument in methods
        if not args[0].is_connected():
            message = "Your connection is already closed"
            logger.error(message, exc_info=False)
            raise OperationOnClosedConnectionError(message)
        return func(*args, **kwargs)

    return wrapped


class Connection(ABC):
    """
    Parent for connection objects, contains all public methods to work with API
    """

    def __init__(self, db_params: Optional[Dict] = None, credentials: Tuple[str, str] = None,
                 timeout: int = DEFAULT_TIMEOUT):
        """
        Init any connection to use
        :param db_params: optional parameter, if it is not None, should be like {"NS": "test", "DB": "test"}
        :param credentials: optional pair of user and pass for auth, like ("root", "root")
        :param timeout: timeout in seconds to wait connection results and responses
        """
        self._db_params = db_params
        self._credentials = credentials
        self._connected = False
        self._timeout = timeout

    def close(self):
        """
        Closes the connection. You can not and should not use a connection object after that
        """
        logger.info("The connection was closed")
        self._connected = False

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self.close()

    def is_connected(self) -> bool:
        """
        Checks the connection is still alive and usable

        :return: True if connection is usable, False otherwise
        """
        return self._connected

    def _in_out_json(self, data, is_loads: bool):
        try:
            return json.loads(data) if is_loads else json.dumps(data)
        except RecursionError as e:
            logger.error("Cant serialize/deserialize object, too many nested levels")
            raise TooManyNestedLevelsError(f"Cant serialize object, too many nested levels\nRefer to: {LINK}") from e

    @connected
    def count(self, table_name: str) -> SurrealResult:
        """
        Returns records count for given table. You should have permissions for this action.
        Actually converts to QL "SELECT count() FROM {table_name} GROUP ALL;" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/functions/count

        Note: returns zero if table does not exist, if you need to check table existence use **is_table_exists**

        Note: if you specify table_name with recordID like "person:john" you will get count of fields in record

        :param table_name: name of the table
        :return: result containing count, like SurrealResult(id='', error=None, result=[{'count': 1}], time='123.333µs')
        """
        logger.info("Query-Operation: COUNT. Table: %s", crop_data(table_name))
        result = self.query(f"SELECT count() FROM {table_name} GROUP ALL;")
        if not result.is_error():
            result.result = self._get_count(result.result)
        return result

    @connected
    def table_info(self, table_name: str) -> SurrealResult:
        """
        Returns info about specified table. You should have permissions for this action.

        Actually converts to QL "INFO FOR TABLE table_name" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :return: full table information
        """
        return self._info(f"TABLE {table_name}")

    @connected
    def db_info(self) -> SurrealResult:
        """
        Returns info about a current database. You should have permissions for this action.

        Actually converts to QL "INFO FOR DB" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :return: full database information
        """
        return self._info("DB")

    @connected
    def ns_info(self) -> SurrealResult:
        """
        Returns info about current namespace. You should have permissions for this action.

        Actually converts to QL "INFO FOR NS" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :return: full namespace information
        """
        return self._info("NS")

    @connected
    def root_info(self) -> SurrealResult:
        """
        Returns info about root. You should have permissions for this action.

        Actually converts to QL "INFO FOR ROOT" to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/info

        :return: information about root
        """
        return self._info("ROOT")

    def _info(self, type_: str) -> SurrealResult:
        logger.info("Query-Operation: %s_INFO", type_)
        return self.query(f"INFO FOR {type_};")

    @connected
    def session_info(self) -> SurrealResult:
        """
        Returns info about the current session. You should have permissions for this action.

        Actually converts to QL query to use in **query** method.

        Refer to: https://docs.surrealdb.com/docs/surrealql/functions/session

        :return: full session information
        """
        query = 'return {"db" : session::db(), "session_id" : session::id(), "ip" : session::ip(), ' \
                '"ns" : session::ns(), "http_origin" : session::origin(), "scope" : session::sc()};'
        logger.info("Query-Operation: SESSION_INFO")
        return self.query(query)

    @connected
    def db_tables(self) -> SurrealResult:
        """
        Returns all tables names in the current database. You should have permissions for this action.

        Actually call **db_info** and parse tables attribute there.

        :return: list of all tables names
        """
        logger.info("Query-Operation: DB_TABLES")
        res: SurrealResult = self.db_info()
        res.result = list(res.result["tables"].keys())
        return res

    @connected
    def is_table_exists(self, table_name: str) -> bool:
        """
        Returns True if table with given name exists in a current database. You should have permissions for this action.

        :param table_name: name of the table, we do not expect record_id here
        :return: True if table exists, False otherwise
        """
        return table_name in self.db_tables().result

    @connected
    def remove_table(self, table_name: str) -> SurrealResult:
        """
        Fully removes table, even if it contains some records, analog of SQL "DROP table". You should have permissions
        for this action.This method cannot remove any other resource, if you need to remove db, ns or scope -
        use **query**

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/remove

        Note: only name of the table allowed here, do not use record_id

        :param table_name: name of the table
        :return: result of the query
        """
        logger.info("Query-Operation: REMOVE. Table name %s", table_name)
        return self.query(f"REMOVE TABLE {table_name} IF EXISTS;")

    @connected
    def show_changes(self, table_name: str, since: str, limit: int = 10) -> SurrealResult:
        """
        This method can use Changes Feed feature of SurrealDB. You need a predefined CHANGEFEED table and database for
        using this action.

        Refer to: https://surrealdb.com/blog/unlocking-streaming-data-magic-with-surrealdb-live-queries-and-change-feeds

        Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#change-feeds

        :param table_name: name of the table, no record_id expected here
        :param since: str representation of ISO date-time, for example "2024-02-06T10:48:08.700483Z"
        :param limit: amount of changes to get
        :return: result of the query
        """
        query = f'SHOW CHANGES FOR TABLE {table_name} SINCE "{since}" LIMIT {limit};'
        return self.query(query)

    @abstractmethod
    def use(self, namespace: str, database: str) -> SurrealResult:
        """
        This method specifies the namespace and database for the current connection

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/use
        """

    @abstractmethod
    def authenticate(self, token: str) -> SurrealResult:
        """
        This method authenticates user with given token.
        Http connections cannot use it
        """

    @abstractmethod
    def invalidate(self) -> SurrealResult:
        """
        This method will invalidate the user's session for the current connection
        Http connections cannot use it
        """

    @abstractmethod
    def let(self, name: str, value) -> SurrealResult:
        """
        This method sets and stores a value which can then be used in a subsequent query

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/let
        """

    @abstractmethod
    def unset(self, name: str) -> SurrealResult:
        """
        This method unsets value, which was previously stored
        """

    @abstractmethod
    def live(self, table_name: str, callback: Callable[[Dict], Any], return_diff: bool = False) -> SurrealResult:
        """
        This method can be used to initiate live query - a real-time selection from a table. Works only for websockets.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/live-select

        About DIFF refer to: https://jsonpatch.com

        Please see surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query
        """

    @abstractmethod
    def custom_live(self, custom_query: str, callback: Callable[[Dict], Any]) -> SurrealResult:
        """
        This method can be used to initiate custom live query - a real-time selection from a table with filters and
        other features of Live Query. Works only for websockets.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/live-select

        Please see surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Note: all results, DIFF, formats etc. should be specified in the query itself
        """

    @abstractmethod
    def kill(self, live_query_id: str) -> SurrealResult:
        """
        This method is used to terminate a running live query by id

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/kill
        """

    @abstractmethod
    def signup(self, namespace: str, database: str, scope: str, params: Optional[Dict] = None) -> SurrealResult:
        """
        This method allows you to sign up a user

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/scope
        """

    @abstractmethod
    def signin(self, user: str, password: str, namespace: Optional[str] = None,
               database: Optional[str] = None, scope: Optional[str] = None) -> SurrealResult:
        """
        This method allows you to sign in a root, namespace, database or scope user against SurrealDB
        """

    @abstractmethod
    def select(self, table_name: str, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method selects either all records in a table or a single record

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/select

        Notice: do not specify id twice: in table name and in record_id, it will cause error on SurrealDB side
        """

    @abstractmethod
    def create(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method creates a record either with a random or specified record_id. If no id specified in record_id or
        in data arguments, then id will be generated by SurrealDB

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/create

        Notice: do not specify id twice, for example, in table name and in data, it will cause error on SurrealDB side
        """

    @abstractmethod
    def insert(self, table_name: str, data: Union[Dict, List]) -> SurrealResult:
        """
        This method inserts one or more records. If you specify recordID in data and record with that id already
        exists - no inserts or updates will happen and the content of the existing record will be returned. If you need
        to change existing record, please consider **update** or **merge**

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/insert

        Note: do not use record id in table_name parameter (table:recordID) - it will cause error on SurrealDB side
        """

    @abstractmethod
    def update(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method can be used to update or create record in the database. So all old fields will be deleted, and new
        will be added, if you wand just to add field to record, keeping old ones -use **merge** method instead. If
        a record with specified id does not exist, it will be created, if exist - all fields will be replaced

        Note: if you want to create/replace one record, you should specify recordID in table_name or in record_id, but
        not in data parameters.

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/update
        """

    @abstractmethod
    def merge(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method merges specified data into either all records in a table or a single record. Old fields in records
        will not be deleted if you want to replace old data with new - use **update** method. If a record with
        specified id does not exist, it will be created.
        """

    @abstractmethod
    def delete(self, table_name: str, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method deletes all records in a table or a single record, be careful and don't forget to specify id if you
        do not want to delete all records. This method does not remove table itself, only records in it.
        As a result of this method, you will get all deleted records or None if no such record or table

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/delete
        """

    @abstractmethod
    def patch(self, table_name: str, data: Union[Dict, List], record_id: Optional[str] = None,
              return_diff: bool = False) -> SurrealResult:
        """
        This method changes specified data in one ar all records. If given table does not exist, new table and record
        will not be created if table exists but no such record_id - new record will be created, if no record id-all
        records will be transformed

        About allowed data format and DIFF refer to: https://jsonpatch.com

        Notice: do not specify id twice, for example, in table name and in record_id, it will cause error
        """

    @abstractmethod
    def query(self, query: str, variables: Optional[Dict] = None) -> SurrealResult:
        """
        This method used for execute a custom SurrealQL query

        For SurrealQL refer to: https://docs.surrealdb.com/docs/surrealql/overview
        """

    @abstractmethod
    def import_data(self, path) -> SurrealResult:
        """
        This method imports a SurrealQL script file into a local or remote SurrealDB database server.

        Note: websocket connection cannot use this method

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#import
        https://docs.surrealdb.com/docs/cli/import
        """

    @abstractmethod
    def export(self) -> str:
        """
        This method exports all data for a specific namespace and database.

        Note: websocket connection cannot use this method

        Refer to: https://docs.surrealdb.com/docs/integration/http#export

        Refer to: https://docs.surrealdb.com/docs/cli/export/
        :return:
        """

    @abstractmethod
    def ml_import(self, path) -> SurrealResult:
        """
        This method imports a SurrealQL ML file into a local or remote SurrealDB database server.

        Note: websocket connection cannot use this method

        Refer to:
        https://docs.surrealdb.com/docs/integration/http#ml-import
        https://docs.surrealdb.com/docs/cli/ml/import
        :param path:
        :return:
        """

    @abstractmethod
    def ml_export(self, name: str, version: str) -> str:
        """
        This method exports a SurrealML machine learning model from a specific namespace and database.
        As machine learning files can be large, the endpoint outputs a chunked HTTP response.

        Note: websocket connection cannot use this method

        Refer to: https://docs.surrealdb.com/docs/integration/http#ml-export

        Refer to: https://docs.surrealdb.com/docs/cli/ml/export
        """

    def _get_count(self, res) -> int:
        if not res:
            return 0
        if 'count' in res:
            return res['count']
        if 'result' in res:
            return self._get_count(res["result"])
        return self._get_count(res[0])



from typing import List

from surrealist.connections import Connection
from surrealist.ql.statements.statement import Statement
from surrealist.utils import OK


class Return(Statement):
    """
    Represent RETURN statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/return

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    """

    def __init__(self, connection: Connection, query: str):
        """
        Init for Create
        :param connection: connection relied on
        :param query: raw query to return
        """
        super().__init__(connection)
        self._query = query

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        return f"RETURN {self._query}"



import urllib.parse
from logging import getLogger
from pathlib import Path
from typing import Optional, Tuple, Dict, Union, List, Callable, Any

from surrealist.clients.ws_client import WebSocketClient
from surrealist.connections.connection import Connection, connected
from surrealist.errors import (SurrealConnectionError, WebSocketConnectionError, ConnectionParametersError,
                               WebSocketConnectionClosedError, CompatibilityError)
from surrealist.result import SurrealResult
from surrealist.utils import DEFAULT_TIMEOUT, crop_data, mask_pass

logger = getLogger("websocket_connection")


class WebSocketConnection(Connection):
    """
    Represents websocket transport and abilities to work with SurrealDb. It is a recommended connection.

    Refer to surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#transports
    Refer to: https://docs.surrealdb.com/docs/integration/websocket

    Each object creates only one websocket connection and can be used in the context manager to close properly.
    You cannot and should not try to use this object after closing connection. Just create a new connection.

    On creating, this object tries to create connection with specified parameters and will raise exception on fail.
    If namespace and database specified - USE method are called automatically
    If credentials specified - SIGNIN are called automatically
    """

    def __init__(self, url: str, db_params: Optional[Dict] = None, credentials: Optional[Tuple[str, str]] = None,
                 timeout: int = DEFAULT_TIMEOUT):
        super().__init__(db_params, credentials, timeout)
        base_url = urllib.parse.urlparse(url.lower())
        self._params = {}
        if db_params:
            if "NS" in db_params:
                self._params["NS"] = db_params["NS"]
            if "DB" in db_params:
                self._params["DB"] = db_params["DB"]
            if not self._params:
                message = "Connection parameters namespace and database required"
                logger.error(message)
                raise ConnectionParametersError(message)
        if base_url.scheme in ("ws", "wss"):
            self._base_url = url
        if base_url.scheme in ("http", "https"):
            self._base_url = f"{base_url.scheme.replace('http', 'ws')}://{base_url.netloc}/rpc"
        try:
            self._client = WebSocketClient(self._base_url, timeout)
        except TimeoutError:
            logger.error("Cant connect to %s in %s seconds", self._base_url, self._timeout)
            raise SurrealConnectionError(f"Cant connect to {self._base_url} in {timeout} seconds.\n"
                                         f"Is your SurrealDB started and work on that url? "
                                         f"Refer to https://docs.surrealdb.com/docs/introduction/start")
        except WebSocketConnectionClosedError:
            logger.error("Cant connect to %s, connection refused", self._base_url)
            raise SurrealConnectionError(f"Cant connect to {self._base_url}, connection refused.\n"
                                         f"Is your SurrealDB started and work on that url? "
                                         f"Refer to https://docs.surrealdb.com/docs/introduction/start")
        self._use_or_sign_on_params(credentials)
        self._connected = True
        masked_creds = None if not credentials else (credentials[0], "******")
        logger.info("Connected to %s, params: %s, credentials: %s, timeout: %s", self._base_url, db_params,
                    masked_creds, timeout)

    def _use_or_sign_on_params(self, credentials):
        if self._params:
            if credentials:
                self._user, self._pass = self._credentials
                signin_result = self.signin(self._user, self._pass, *list(self._params.values()))
                if signin_result.is_error():
                    logger.error("Error on connecting to %s. Info %s", self._base_url, signin_result)
                    raise WebSocketConnectionError(f"Error on connecting to {self._base_url}.\n"
                                                   f"Info: {signin_result.result}")
            else:
                use_result = self.use(self._params["NS"], self._params["DB"])
                if use_result.is_error():
                    logger.error("Error on use %s. Info %s", self._params, use_result)
                    raise WebSocketConnectionError(f"Error on use '{self._params}'.\nInfo: {use_result.result}")
        else:
            if credentials:
                self._user, self._pass = self._credentials
                signin_result = self.signin(self._user, self._pass)
                if signin_result.is_error():
                    logger.error("Error on connecting to %s. Info %s", self._base_url, signin_result)
                    raise WebSocketConnectionError(f"Error on connecting to '{self._base_url}'.\nInfo: {signin_result}")

    @connected
    def use(self, namespace: str, database: str) -> SurrealResult:
        """
        This method specifies the namespace and database for the current connection

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#use

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/use

        Example:
        websocket_connection.use("test", "test") # use test namespace and test database for this connection

        :param namespace: name of the namespace to use
        :param database: name of the database to use
        :return: result of request
        """
        data = {"method": "use", "params": [namespace, database]}
        logger.info("Operation: USE. Namespace: %s, database %s", crop_data(namespace), crop_data(database))
        result = self._run(data)
        if not result.is_error():
            # if USE was OK we need to store new data (ns and db)
            self._params = {"NS": namespace, "DB": database}
        return result

    @connected
    def signin(self, user: str, password: str, namespace: Optional[str] = None, database: Optional[str] = None,
               scope: Optional[str] = None) -> SurrealResult:
        """
        This method allows you to sign in a root, namespace, database or scope user against SurrealDB

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#signin

        Example:
        websocket_connection.signin('root', 'root') # sign in as root user
        websocket_connection.signin('root', 'root', namespace='test', database='test') # sign in as db user

        :param user: name of the user
        :param password: password for auth
        :param namespace: name of the namespace to use
        :param database: name of the database to use
        :param scope: name of the scope to use
        :return: result of request
        """
        params = {"user": user, "pass": password}
        if namespace is not None:
            params["NS"] = namespace
        if database is not None:
            params["DB"] = database
        if scope is not None:
            params["SC"] = scope
        data = {"method": "signin", "params": [params]}
        logger.info("Operation: SIGNIN. Data: %s", crop_data(mask_pass(str(params))))
        return self._run(data)

    @connected
    def signup(self, namespace: str, database: str, scope: str, params: Optional[Dict] = None) -> SurrealResult:
        """
        This method allows you to sign up a user

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#signup

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/scope

        Example:
        websocket_connection.signup(namespace='test', database='test', scope='user_scope',
                              params={'user': 'john:doe', 'pass': '123456'})

        :param namespace: name of the namespace to use
        :param database: name of the database to use
        :param scope: name of the scope to use
        :param params: dict with user and pass, for example {"user":"root", "pass":"root"}
        :return: result of request
        """
        params = params or {}
        params = {"NS": namespace, "DB": database, "SC": scope, **params}
        data = {"method": "signup", "params": [params]}
        logger.info("Operation: SIGNUP. Data: %s", crop_data(mask_pass(str(params))))
        return self._run(data)

    @connected
    def authenticate(self, token: str) -> SurrealResult:
        """
        This method authenticates user with given token

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#authenticate

        :param token: token for auth
        :return: result of request
        """
        data = {"method": "authenticate", "params": [token]}
        logger.info("Operation: AUTHENTICATE. Token: %s", crop_data(token))
        return self._run(data)

    @connected
    def invalidate(self) -> SurrealResult:
        """
        This method will invalidate the user's session for the current connection

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#invalidate

        :return: result of request
        """
        data = {"method": "invalidate"}
        logger.info("Operation: INVALIDATE")
        return self._run(data)

    @connected
    def let(self, name: str, value: Any) -> SurrealResult:
        """
        This method sets and stores a value which can then be used in a subsequent query

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#let

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/let

        :param name: name for the variable (without $ sign!)
        :param value: value for the variable
        :return: result of request
        """
        data = {"method": "let", "params": [name, value]}
        logger.info("Operation: LET. Name: %s, Value: %s", crop_data(name), crop_data(str(value)))
        return self._run(data)

    @connected
    def unset(self, name: str) -> SurrealResult:
        """
        This method unsets value, which was previously stored

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#unset

        :param name: name for the variable (without $ sign!)
        :return: result of request
        """
        data = {"method": "unset", "params": [name]}
        logger.info("Operation: UNSET. Variable name: %s", crop_data(name))
        return self._run(data)

    @connected
    def live(self, table_name: str, callback: Callable[[Dict], Any], return_diff: bool = False) -> SurrealResult:
        """
        This method can be used to initiate live query - a real-time selection from a table

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#live

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/live-select

        About DIFF refer to: https://jsonpatch.com

        Please see surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Example:
        websocket_connection.live("article", callback=lambda a_dict: print(a_dict)) # creates a live query to check any
        events on article table and just printing all incoming results

        :param table_name: name of the table to observe
        :param callback: a function to call on any incoming event. It should take one argument - a dict
        :param return_diff: True if you want to get only DIFF info on table events, False for a standard results
        :return: result of request with the live_id in 'result' field
        """
        params = [table_name]
        if return_diff:
            params.append(True)
        data = {"method": "live", "params": params}
        logger.info("Operation: LIVE. Data: %s", crop_data(str(params)))
        return self._run(data, callback)

    @connected
    def custom_live(self, custom_query: str, callback: Callable[[Dict], Any]) -> SurrealResult:
        """
        This method can be used to initiate custom live query - a real-time selection from a table with filters and
        other features of Live Query

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/live-select

        Please see surrealist documentation: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Note: all results, DIFF, formats etc. should be specified in the query itself

        Example:
        ws.custom_live("LIVE SELECT * FROM person WHERE age > 18;", callback=lambda a_dict: print(a_dict)) # creates
        a live query to check any events on person table with records, where age field is bigger than 18, and just
        printing all incoming results

        :param custom_query: full LIVE SELECT query text
        :param callback: a function to call on any incoming event. It should take one argument - a dict
        :return: result of request with the live_id in 'result' field
        """
        data = {"method": "query", "params": [custom_query], "additional": "live"}
        logger.info("Operation: CUSTOM LIVE. Query: %s", crop_data(custom_query))
        result = self._run(data, callback)
        result.query = custom_query
        return result

    @connected
    def kill(self, live_query_id: str) -> SurrealResult:
        """
        This method is used to terminate a running live query by id

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#kill

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/kill

        :param live_query_id: id for the query to kill
        :return: result of request
        """
        data = {"method": "kill", "params": [live_query_id]}
        logger.info("Operation: KILL. Live_id: %s", crop_data(live_query_id))
        return self._run(data)

    @connected
    def query(self, query: str, variables: Optional[Dict] = None) -> SurrealResult:
        """
        This method used for execute a custom SurrealQL query

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#query

        For SurrealQL refer to: https://docs.surrealdb.com/docs/surrealql/overview

        Example:
        websocket_connection.query("SELECT * FROM article;") # gets all records from article table
        websocket_connection.query("SELECT * FROM type::table($tb);", {"tb": "article"}) # gets all records from
        article table using variable tb to specify table

        :param query: any SurrealQL query to execute
        :param variables: a set of variables used by the query
        :return: result of request
        """
        params = [query]
        if variables is not None:
            params.append(variables)
        data = {"method": "query", "params": params}
        logger.info("Operation: QUERY. Query: %s, variables: %s", crop_data(query), crop_data(str(variables)))
        result = self._run(data)
        result.query = params[0] if len(params) == 1 else params
        return result

    @connected
    def select(self, table_name: str, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method selects either all records in a table or a single record

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#select

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/select

        Examples:
        websocket_connection.select("article") # select all records in article table
        websocket_connection.select("article:first") # select one article with id 'first'
        websocket_connection.select("article", "first") # select one article with id 'first', analog of previous

        Notice: do not specify id twice: in table name and in record_id, it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to select
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        table_name = table_name if record_id is None else f"{table_name}:{record_id}"
        data = {"method": "select", "params": [table_name]}
        logger.info("Operation: SELECT. Path: %s", crop_data(table_name))
        result = self._run(data)
        if not isinstance(result.result, List) and not result.is_error():
            result.result = [result.result] if result.result else []
        return result

    @connected
    def create(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method creates a record either with a random or specified record_id. If no id specified in record_id or
        in data arguments, then id will be generated by SurrealDB

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#create

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/create

        Examples:
        websocket_connection.create("person", {"name": "John Doe"}) # create one record in person table with random id
        websocket_connection.create("person", {"id":"my_id", "name": "John Doe"}) # create one record in person table
        with specified id
        websocket_connection.create("person", {"name": "John Doe"}, "my_id") # create one record in person table
        with specified id
        websocket_connection.create("person:my_id", {"name": "John Doe"}) # create one record in person table
        with specified id

        Notice: do not specify id twice, for example, in table name and in data, it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to create
        :param data: dict with data to create
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        if record_id is not None:
            data["id"] = record_id
        _data = {"method": "create", "params": [table_name, data]}
        logger.info("Operation: CREATE. Path: %s, data: %s", crop_data(table_name), crop_data(str(data)))
        result = self._run(_data)
        if isinstance(result.result, List) and len(result.result) == 1:
            result.result = result.result[0]
        return result

    @connected
    def insert(self, table_name: str, data: Union[List, Dict]) -> SurrealResult:
        """
        This method inserts one or more records. If you specify recordID in data and record with that id already
        exists - no inserts or updates will happen and the content of the existing record will be returned. If you need
        to change existing record, please consider **update** or **merge**

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#insert

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/insert

        Examples:
        websocket_connection.insert("person", {"name": "John Doe"}) # inserts one record with random id
        websocket_connection.insert("person", [{"name": "John Doe"}, {"name", "Jane Doe"}]) # inserts two records
        with random ids

        Note: do not use record id in table_name parameter (table:recordID) - it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to insert
        :param data: dict or list(many records) with data to create
        :return: result of request
        """
        _data = {"method": "insert", "params": [table_name, data]}
        logger.info("Operation: INSERT. Path: %s, data: %s", crop_data(table_name), crop_data(str(data)))
        return self._run(_data)

    @connected
    def update(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method can be used to update or modify records in the database. So all old fields will be deleted, and new
        will be added, if you wand just to add field to record, keeping old ones -use **merge** method instead.If
        record with specified id does not exist it will be created, if it exists - all fields will be replaced

        Note: if you want to create/replace one record, you should specify recordID in table_name or in record_id, but
        not in data parameters.

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#update

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/update

        Example:
        websocket_connection.update("person:my_id", {"name": "Alex Doe"}) # record with specified id will be now
        {"name": "Alex Doe"}, all data stored in record before will be deleted
        websocket_connection.update("person", {"name": "Alex Doe"}, "my_id") # record with specified id will be now
        {"name": "Alex Doe"}, all data stored in record before will be deleted

        Notice: do not specify id twice, for example, in table name and in record_id, it will cause error

        :param table_name: table name or table name with record_id to update
        :param data: dict with data to create
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        table_name = table_name if record_id is None else f"{table_name}:{record_id}"
        _data = {"method": "update", "params": [table_name, data]}
        logger.info("Operation: UPDATE. Path: %s, data: %s", crop_data(table_name), crop_data(str(data)))
        return self._run(_data)

    @connected
    def merge(self, table_name: str, data: Dict, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method merges specified data into either all records in a table or a single record. Old data in records
        will not be deleted, if you want to replace old data with new - use **update** method.If
        record with specified id does not exist, it will be created.

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#merge

        Examples:
        websocket_connection.merge("person",{"active": True}) # "active" will be added to all records in person table
        websocket_connection.merge("person:my_id", {"active": True}) # "active" will be added to one record in person
        table with specified id

        :param table_name: table name or table name with record_id to merge
        :param data: dict with data to add
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        table_name = table_name if record_id is None else f"{table_name}:{record_id}"
        _data = {"method": "merge", "params": [table_name, data]}
        logger.info("Operation: MERGE. Path: %s, data: %s", crop_data(table_name), crop_data(str(data)))
        return self._run(_data)

    @connected
    def patch(self, table_name: str, data: Union[Dict, List], record_id: Optional[str] = None,
              return_diff: bool = False) -> SurrealResult:
        """
        This method changes specified data in one ar all records. If given table does not exist, new table and record
        will not be created if table exists but no such record_id - new record will be created, if no record id-all
        records will be transformed

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#patch

        About allowed data format and DIFF refer to: https://jsonpatch.com

        Examples:
        websocket_connection.patch("person", [{"op": "replace", "path": "/active", "value": False}]) # replaces active
        field for all records in person table to False
        websocket_connection.patch("person:my_id", [{"op": "replace", "path": "/active", "value": False}]) # replaces
        active field for one record with specified id to False

        Notice: do not specify id twice, for example, in table name and in data, it will cause error on SurrealDB side

        :param table_name: table name or table name with record_id to patch
        :param data: list with json-patch data
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :param return_diff: True if you want to get only DIFF info, False for a standard results
        :return: result of request
        """
        table_name = table_name if record_id is None else f"{table_name}:{record_id}"
        params = [table_name, data]
        if return_diff:
            params.append(return_diff)
        _data = {"method": "patch", "params": params}
        logger.info("Operation: PATCH. Path: %s, data: %s, use DIFF: %s", crop_data(table_name), crop_data(str(data)),
                    return_diff)
        return self._run(_data)

    @connected
    def delete(self, table_name: str, record_id: Optional[str] = None) -> SurrealResult:
        """
        This method deletes all records in a table or a single record, be careful and don't forget to specify id if you
        do not want to delete all records. This method does not remove table itself, only records in it.As a result of
        this method you will get all deleted records or None if no such record or table

        Refer to: https://docs.surrealdb.com/docs/integration/websocket#delete

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/delete

        Examples:
        websocket_connection.delete("person:my_id") # deletes one record in person table
        websocket_connection.delete("person", "my_id") # deletes one record in person table
        websocket_connection.delete("person") # deletes all records in person table

        :param table_name: table name or table name with record_id to delete
        :param record_id: optional parameter, if it exists it will transform table_name to "table_name:record_id"
        :return: result of request
        """
        table_name = table_name if record_id is None else f"{table_name}:{record_id}"
        _data = {"method": "delete", "params": [table_name]}
        logger.info("Operation: DELETE. Path: %s", crop_data(table_name))
        return self._run(_data)

    def export(self):
        """
        Websocket transport cannot use export operation, so you can use http transport for that, or SurrealDB tools

        Refer to: https://docs.surrealdb.com/docs/cli/export

        :raise CompatibilityError: on any use
        """
        message = "Export is not allowed for websocket transport in the current SurrealDB version"
        logger.error(message)
        full_message = f"{message}\nYou can use http transport or abilities of SurrealDb itself\n" \
                       f"Refer to: https://docs.surrealdb.com/docs/cli/export"
        raise CompatibilityError(full_message)

    def ml_export(self, _name: str, _version: str):
        """
        Websocket transport cannot use ML export operation, so you can use http transport for that, or SurrealDB tools

        Refer to: https://docs.surrealdb.com/docs/cli/ml/export

        :raise CompatibilityError: on any use
        """
        message = "ML export is not allowed for websocket transport in the current SurrealDB version"
        logger.error(message)
        full_message = f"{message}\nYou can use http transport or abilities of SurrealDb itself\n" \
                       f"Refer to: https://docs.surrealdb.com/docs/cli/ml/export"
        raise CompatibilityError(full_message)

    def import_data(self, _path: Union[str, Path]):
        """
        Websocket transport cannot use import operation, so you can use http transport for that, or SurrealDB tools

        Refer to: https://docs.surrealdb.com/docs/cli/import

        :raise CompatibilityError: on any use
        """
        message = "Import is not allowed for websocket transport in the current SurrealDB version"
        logger.error(message)
        full_message = f"{message}\nYou can use http transport or abilities of SurrealDb itself\n" \
                       f"Refer to: https://docs.surrealdb.com/docs/cli/import"
        raise CompatibilityError(full_message)

    def ml_import(self, _path: Union[str, Path]):
        """
        Websocket transport cannot use ML import operation, so you can use http transport for that, or SurrealDB tools

        Refer to: https://docs.surrealdb.com/docs/cli/ml/import

        :raise CompatibilityError: on any use
        """
        message = "ML import is not allowed for websocket transport in the current SurrealDB version"
        logger.error(message)
        full_message = f"{message}\nYou can use http transport or abilities of SurrealDb itself\n" \
                       f"Refer to: https://docs.surrealdb.com/docs/cli/ml/import"
        raise CompatibilityError(full_message)

    def close(self):
        super().close()
        if self.is_connected():
            self._client.close()

    def is_connected(self) -> bool:
        return self._client.is_connected()

    def _run(self, data, callback: Callable = None) -> SurrealResult:
        result = self._client.send(data, callback)
        logger.info("Got result: %s", crop_data(str(result)))
        return result



from typing import Tuple

from surrealist.ql.statements.statement import Statement, FinishedStatement, IterableStatement
from surrealist.utils import OK


class Explain(FinishedStatement):
    """ Explain is not a child of IterableStatement because it has always only one result"""

    def __init__(self, statement: Statement, full: bool = False):
        super().__init__(statement)
        self._full = full

    def _clean_str(self):
        final = " FULL" if self._full else ""
        return f"{self._statement._clean_str()} EXPLAIN{final}"


class SelectUseExplain:

    def explain(self) -> Explain:
        return Explain(self, full=False)

    def explain_full(self) -> Explain:
        return Explain(self, full=True)


class Parallel(IterableStatement, SelectUseExplain):

    def _clean_str(self):
        return f"{self._statement._clean_str()} PARALLEL"


class SelectUseParallel(SelectUseExplain):
    def parallel(self) -> Parallel:
        return Parallel(self)


class Timeout(IterableStatement, SelectUseParallel):
    def __init__(self, statement: Statement, duration: str):
        super().__init__(statement)
        self._duration = duration

    def _validate(self):
        if not any(self._duration.endswith(letter) for letter in ('ms', 's', 'm')):
            return f"Wrong duration {self._duration}, allowed postfix are (ms, s, m)"
        if ' ' in self._duration:
            return "Wrong duration format, should be like 5s"
        return OK

    def _clean_str(self):
        return f"{self._statement._clean_str()} TIMEOUT {self._duration}"


class SelectUseTimeout(SelectUseParallel):
    def timeout(self, duration: str) -> Timeout:
        return Timeout(self, duration)


class Fetch(IterableStatement, SelectUseTimeout):
    def __init__(self, statement: Statement, *args: str):
        super().__init__(statement)
        self._args = args

    def _clean_str(self):
        add = "" if not self._args else ", ".join(self._args)
        if not add:
            return self._statement
        return f"{self._statement._clean_str()} FETCH {add}"


class SelectUseFetch(SelectUseTimeout):
    def fetch(self, *args) -> Fetch:
        return Fetch(self, *args)


class Start(IterableStatement, SelectUseFetch):
    def __init__(self, statement: Statement, offset: int):
        super().__init__(statement)
        self._offset = offset

    def _validate(self):
        if self._offset < 0:
            return "Offset should not be negative number"
        return OK

    def _clean_str(self):
        return f"{self._statement._clean_str()} START {self._offset}"


class SelectUseStart(SelectUseFetch):
    def start_at(self, offset: int) -> Start:
        return Start(self, offset)


class Limit(IterableStatement, SelectUseStart):
    def __init__(self, statement: Statement, limit: int):
        super().__init__(statement)
        self._limit = limit

    def _validate(self):
        if self._limit < 1:
            return "Limit should not be less than 1"
        return OK

    def _clean_str(self):
        return f"{self._statement._clean_str()} LIMIT {self._limit}"


class SelectUseLimit(SelectUseStart):
    def limit(self, limit: int) -> Limit:
        return Limit(self, limit)


class OrderByRand(IterableStatement, SelectUseLimit):
    def __init__(self, statement: Statement):
        super().__init__(statement)
        self._statement = statement

    def _clean_str(self):
        return f"{self._statement._clean_str()} ORDER BY RAND()"


class OrderBy(IterableStatement, SelectUseLimit):
    def __init__(self, statement: Statement, *args: str):
        super().__init__(statement)
        self._statement = statement
        self._args = args

    def _clean_str(self):
        if not self._args:
            return self._statement
        what = ", ".join(self._args)
        return f"{self._statement._clean_str()} ORDER BY {what}"


class SelectUseOrder(SelectUseLimit):
    def order_by(self, *args: str) -> OrderBy:
        return OrderBy(self, *args)

    def order_by_rand(self) -> OrderByRand:
        return OrderByRand(self)


class GroupBy(IterableStatement, SelectUseOrder):
    def __init__(self, statement: Statement, *args):
        super().__init__(statement)
        self._statement = statement
        self._args = args

    def _clean_str(self):
        if not self._args:
            return self._statement
        what = ", ".join(self._args)
        return f"{self._statement._clean_str()} GROUP BY {what}"


class GroupAll(IterableStatement, SelectUseOrder):
    def __init__(self, statement: Statement):
        super().__init__(statement)
        self._statement = statement

    def _clean_str(self):
        return f"{self._statement._clean_str()} GROUP ALL"


class SelectUseGroup(SelectUseOrder):
    def group_by(self, *args: str) -> GroupBy:
        return GroupBy(self, *args)

    def group_all(self) -> GroupAll:
        return GroupAll(self)


class Split(IterableStatement, SelectUseGroup):
    def __init__(self, statement: Statement, field: str):
        super().__init__(statement)
        self._statement = statement
        self._field = field

    def _clean_str(self):
        return f"{self._statement._clean_str()} SPLIT {self._field}"


class SelectUseSplit(SelectUseGroup):
    def split(self, field: str) -> Split:
        return Split(self, field)


class Or(IterableStatement, SelectUseSplit):
    def __init__(self, statement: Statement, predicate: str):
        super().__init__(statement)
        self._statement = statement
        self._predicate = predicate

    def AND(self, predicate):
        return And(self, predicate)

    def OR(self, predicate):
        return Or(self, predicate)

    def _clean_str(self):
        return f"{self._statement._clean_str()} OR {self._predicate}"


class And(IterableStatement, SelectUseSplit):
    def __init__(self, statement: Statement, predicate: str):
        super().__init__(statement)
        self._statement = statement
        self._predicate = predicate

    def AND(self, predicate):
        return And(self, predicate)

    def OR(self, predicate):
        return Or(self, predicate)

    def _clean_str(self):
        return f"{self._statement._clean_str()} AND {self._predicate}"


class Where(IterableStatement, SelectUseSplit):
    def __init__(self, statement: Statement, predicate: str):
        super().__init__(statement)
        self._statement = statement
        self._predicate = predicate

    def AND(self, predicate):
        return And(self, predicate)

    def OR(self, predicate):
        return Or(self, predicate)

    def _clean_str(self):
        return f"{self._statement._clean_str()} WHERE {self._predicate}"


class SelectUseWhere(SelectUseSplit):
    def where(self, predicate: str) -> Where:
        return Where(self, predicate)


class WithIndex(IterableStatement, SelectUseWhere):
    def __init__(self, statement: Statement, *index_names: str):
        super().__init__(statement)
        self._statement = statement
        self._index = index_names

    def _clean_str(self):
        if not self._index:
            return self._statement
        indexes = ", ".join(self._index)
        return f"{self._statement._clean_str()} WITH INDEX {indexes}"


class WithNoIndex(IterableStatement, SelectUseWhere):
    def __init__(self, statement: Statement):
        super().__init__(statement)
        self._statement = statement

    def _clean_str(self):
        return f"{self._statement._clean_str()} WITH NO INDEX"


class SelectUseIndex(SelectUseWhere):
    def with_index(self, *args: Tuple[str]) -> WithIndex:
        return WithIndex(self, *args)

    def with_no_index(self) -> WithNoIndex:
        return WithNoIndex(self)



import json
from typing import Dict, List, Optional

from surrealist.ql.statements.common_statements import CanUseWhere
from surrealist.ql.statements.statement import FinishedStatement, Statement


class Set(FinishedStatement, CanUseWhere):
    def __init__(self, statement: Statement, result: Optional[str] = None, **kwargs):
        super().__init__(statement)
        self._kw = kwargs
        self._result = result

    def _clean_str(self):
        if not self._kw and not self._result:
            return self._statement._clean_str()
        if not self._result:
            args = ", ".join(f"{k} = {json.dumps(v)}" for k, v in self._kw.items())
        else:
            args = self._result
        return f"{self._statement._clean_str()} SET {args}"


class Patch(FinishedStatement, CanUseWhere):
    def __init__(self, statement: Statement, operations: List[Dict]):
        super().__init__(statement)
        self._value = operations

    def _clean_str(self):
        return f"{self._statement._clean_str()} PATCH {json.dumps(self._value)}"


class Merge(FinishedStatement, CanUseWhere):
    def __init__(self, statement: Statement, value: Dict):
        super().__init__(statement)
        self._value = value

    def _clean_str(self):
        return f"{self._statement._clean_str()} MERGE {json.dumps(self._value)}"


class Content(FinishedStatement, CanUseWhere):
    def __init__(self, statement: Statement, value: Dict):
        super().__init__(statement)
        self._value = value

    def _clean_str(self):
        return f"{self._statement._clean_str()} CONTENT {json.dumps(self._value)}"


class UpdateUseMethods(CanUseWhere):
    def set(self, result: Optional[str] = None, **kwargs) -> Set:
        return Set(self, result, **kwargs)

    def content(self, value: Dict) -> Content:
        return Content(self, value)

    def merge(self, value: Dict) -> Merge:
        return Merge(self, value)

    def patch(self, operations: List[Dict]) -> Patch:
        return Patch(self, operations)



from .connections import WebSocketConnection, HttpConnection, Connection
from .errors import *
from .ql import Database, Table
from .result import SurrealResult
from .surreal import Surreal
from .utils import DATA_LENGTH_FOR_LOGS, get_uuid

__all__ = ("Surreal", "SurrealResult", "WebSocketConnection", "HttpConnection", "PySurrealError", "HttpConnectionError",
           "HttpClientError", "SurrealConnectionError", "WebSocketConnectionError", "WebSocketConnectionClosedError",
           "ConnectionParametersError", "CompatibilityError", "OperationOnClosedConnectionError", "WrongCallError",
           "DATA_LENGTH_FOR_LOGS", "Connection", "get_uuid", "Database", "Table")



import json
from typing import Dict, Optional

from surrealist.ql.statements.common_statements import CanUseReturn
from surrealist.ql.statements.statement import FinishedStatement, Statement


class Set(FinishedStatement, CanUseReturn):
    def __init__(self, statement: Statement, result: Optional[str] = None, **kwargs):
        super().__init__(statement)
        self._kw = kwargs
        self._result = result

    def _clean_str(self):
        if not self._kw and not self._result:
            return self._statement._clean_str()
        if not self._result:
            args = ", ".join(f"{k} = {json.dumps(v)}" for k, v in self._kw.items())
        else:
            args = self._result
        return f"{self._statement._clean_str()} SET {args}"


class Content(FinishedStatement, CanUseReturn):
    def __init__(self, statement: Statement, content: Dict):
        super().__init__(statement)
        self._content = content

    def _clean_str(self):
        args = json.dumps(self._content)
        return f"{self._statement._clean_str()} CONTENT {args}"


class CreateUseSetContent(CanUseReturn):
    def content(self, content: Dict) -> Content:
        return Content(self, content)

    def set(self, result: Optional[str] = None, **kwargs) -> Set:
        return Set(self, result, **kwargs)



from typing import List

from surrealist import Connection
from surrealist.utils import OK
from .show_statements import ShowUseSince
from .statement import Statement


class Show(Statement, ShowUseSince):
    """
    Represents SHOW CHANGES statement, it should be able to use any statements from documentation

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/show

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_show_examples.py
    """

    def __init__(self, connection: Connection, table_name: str):
        super().__init__(connection)
        self._table_name = table_name

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        return f"SHOW CHANGES FOR TABLE {self._table_name}"



from typing import Optional, Callable, Dict, Tuple, Union

from surrealist import WrongCallError
from surrealist.connections import Connection
from surrealist.ql.statements.create import Create
from surrealist.ql.statements.delete import Delete
from surrealist.ql.statements.insert import Insert
from surrealist.ql.statements.live import Live
from surrealist.ql.statements.remove import Remove
from surrealist.ql.statements.select import Select
from surrealist.ql.statements.show import Show
from surrealist.ql.statements.statement import Statement
from surrealist.ql.statements.update import Update
from surrealist.result import SurrealResult


class Table:
    """
    Represents a table of the database including not (yet) existing one.
    Can use only statements of the table level (CRUD).
    If you need a raw query, DEFINE or transactions - you need database level

    Please refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#methods-and-query-language

    SurrealQL: https://docs.surrealdb.com/docs/surrealql/overview

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/
    """

    def __init__(self, name: str, connection: Connection):
        self._connection = connection
        self._name = name

    @property
    def name(self) -> str:
        """
        Return name of the table

        :return: table name
        """
        return self._name

    def info(self) -> Dict:
        """
        Returns full table info

        :return: Result of the request
        """
        return self._connection.table_info(self._name).result

    def count(self) -> int:
        """
        Returns the number of records at current table, returns 0 if table is empty or not exist

        :return: number of records
        """
        return self._connection.count(self._name).result

    def select(self, *args, alias: Optional[Tuple[str, Union[str, Statement]]] = None,
               value: Optional[str] = None) -> Select:
        """
        Represents SELECT statement and its abilities as refer here:
        https://docs.surrealdb.com/docs/surrealql/statements/select

        Example: table.select('id', 'name').run()

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_select_examples.py

        :param alias: pairs of names and values
        :param value: on exists an add VALUE statement
        :param args: which fields to select, if no fields or "*" - selects all
        :return: Select object
        """
        return Select(self._connection, self.name, *args, alias=alias, value=value)

    def create(self, record_id: Optional[Union[str, int]] = None) -> Create:
        """
        Represent CREATE a statement and its abilities as refer here:
        https://docs.surrealdb.com/docs/surrealql/statements/create

        Example:
        db.table("article").create("first").content({"author": "author:john", "title": uid}).run()

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_create_examples.py

        :param record_id: optional, if specified transform to 'table_name:record_id'
        :return: Create object
        """
        return Create(self._connection, self.name, record_id)

    def show_changes(self) -> Show:
        """
        Represents SHOW CHANGES statement for the Change Feed

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/show

        Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#change-feeds

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_show_examples.py

        :return: Show object
        """
        return Show(self._connection, self._name)

    def delete(self, record_id: Optional[str] = None) -> Delete:
        """
        Represent DELETE statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/delete

        Example:
        db.table("author").delete("john").return_none().run()

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_delete_examples.py

        :param record_id: optional, if specified transform to 'table_name:record_id'
        :return: Delete object
        """
        return Delete(self._connection, self._name, record_id)

    def delete_all(self) -> SurrealResult:
        """
        Deletes all records at the database and returns nothing.
        If you need to get back all data ot id's - use query builder or raw_query on database level.
        This action does not remove table, just records in it.
        If you need to delete table itself - use **drop** method

        :return: result with [] as a response
        """
        return Delete(self._connection, self._name).return_none().run()

    def drop(self) -> SurrealResult:
        """
        Fully removes table with all records in it.
        If you need to just delete all records and keep table - use **delete_all** method

        :return: result of response
        """
        return Remove(self._connection, self._name).run()

    def remove(self) -> SurrealResult:
        """
        Fully removes table with all records in it
        If you need to just delete all records and keep table - use **delete_all** method

        :return: result of response
        """
        return self.drop()

    def live(self, callback: Callable[[Dict], None], use_diff: bool = False) -> Live:
        """
        Represents LIVE statement for a live query

        Example:
        db.person.live(func).alias("first_name", "NAME").where("age > 22").run()

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/live-select

        Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_live_examples.py

        :param callback: function to call on live query event
        :param use_diff: return result in DIFF format
        :return: Live object
        """
        return Live(self._connection, self._name, callback, use_diff)

    def kill(self, live_id: str) -> SurrealResult:
        """
        Represents a KILL statement, for killing a live query

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/kill

        :param live_id: id of the query
        :return: result
        """
        return self._connection.kill(live_id)

    def insert(self, *args) -> Insert:
        """
        Represent INSERT INTO statement.
        Arguments here are:
          - the only one which is Dict or List or Statement

          OR

          - multiple arguments where first are names and other - values

        Example:
        db.person.insert(("name", "age"), ("Tobie", 33)).run()
        db.person.insert({("name":"Tobie", "age": 33)}).run()

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/insert

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_insert_examples.py

        :param args: args for insert, it can be list of records, one record, one statement or 2 or more tuples of names
        and values
        :return: Insert object
        """
        return Insert(self._connection, self._name, *args)

    def update(self, record_id: Optional[str] = None) -> Update:
        """
        Represent UPDATE object

        Example:
        db.table("user").update("alex").only().merge({"active": True}).run()

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/update

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_update_examples.py

        :param record_id: optional, if specified transform to 'table_name:record_id'
        :return: Update object
        """
        return Update(self._connection, self._name, record_id)

    def __repr__(self):
        return f"Table(name={self._name})"

    def __call__(self, *args, **kwargs):
        raise WrongCallError(f"Table object is not callable. \n"
                             f"It looks like you misspelled the method name of Database({self._name})")



from datetime import datetime

from surrealist.ql.statements.statement import FinishedStatement, Statement
from surrealist.utils import OK, DATE_FORMAT, DATE_FORMAT_NS


class Limit(FinishedStatement):
    def __init__(self, statement: Statement, limit: int):
        super().__init__(statement)
        self._limit = limit

    def _validate(self):
        if self._limit < 1:
            return "Limit should not be less than 1"
        return OK

    def _clean_str(self):
        return f"{self._statement._clean_str()} LIMIT {self._limit}"


class ShowUseLimit:
    def limit(self, limit: int) -> Limit:
        return Limit(self, limit)


class Since(FinishedStatement, ShowUseLimit):
    def __init__(self, statement: Statement, timestamp: str):
        super().__init__(statement)
        self._timestamp = timestamp

    def _validate(self):
        try:
            format_ = DATE_FORMAT if "." not in self._timestamp else DATE_FORMAT_NS
            datetime.strptime(self._timestamp, format_)
        except ValueError:
            return "Timestamp in wrong format, you need iso-date like 2024-01-01T10:10:10.000001Z"
        return OK

    def _clean_str(self):
        return f'{self._statement._clean_str()} SINCE "{self._timestamp}"'


class ShowUseSince(ShowUseLimit):
    def since(self, timestamp: str) -> Since:
        return Since(self, timestamp)



from typing import List, Union, Any

from surrealist import Connection
from surrealist.ql.statements.statement import Statement
from surrealist.utils import OK


class DefineEvent(Statement):
    """
    Represents DEFINE EVENT statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/event

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

    """

    def __init__(self, connection: Connection, name: str, table_name: str, then: Union[str, Statement]):
        super().__init__(connection)
        self._name = name
        self._table_name = table_name
        self._when = None
        self._then = then if not isinstance(then, Statement) else f"({then._clean_str()})"

    def when(self, predicate: str) -> "DefineEvent":
        """
        Represents condition for event, WHEN statement

        :param predicate: condition for event
        """
        self._when = predicate
        return self

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        when = f"WHEN {self._when} " if self._when else ""
        return f"DEFINE EVENT {self._name} ON TABLE {self._table_name} {when}THEN {self._then}"


class DefineUser(Statement):
    """
    Represents DEFINE USER statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/user

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    """

    def __init__(self, connection: Connection, user_name: str, password: str):
        super().__init__(connection)
        self._name = user_name
        self._pass = password
        self._role = "VIEWER"

    def role_owner(self) -> "DefineUser":
        """
        Represents role OWNER

        """
        self._role = "OWNER"
        return self

    def role_editor(self) -> "DefineUser":
        """
        Represents role EDITOR

        """
        self._role = "EDITOR"
        return self

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        return f"DEFINE USER {self._name} ON DATABASE PASSWORD '{self._pass}' ROLES {self._role}"


class DefineParam(Statement):
    """
    Represents DEFINE PARAM statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/param

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    """

    def __init__(self, connection: Connection, param_name: str, value: Any):
        super().__init__(connection)
        self._name = param_name
        self._value = value

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        return f"DEFINE PARAM ${self._name} VALUE {self._value}"


class DefineAnalyzer(Statement):
    """
    Represents DEFINE ANALYZER statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/analyzer

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    """

    def __init__(self, connection: Connection, name: str):
        super().__init__(connection)
        self._name = name
        self._tokenizers = None
        self._filters = None

    def tokenizers(self, value: str) -> "DefineAnalyzer":
        """
        Represents TOKENIZERS statement
        """
        self._tokenizers = value
        return self

    def filters(self, value: str) -> "DefineAnalyzer":
        """
        Represents FILTERS statement
        """
        self._filters = value
        return self

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        tok = "" if not self._tokenizers else f" TOKENIZERS {self._tokenizers}"
        filters = "" if not self._filters else f" FILTERS {self._filters}"
        return f"DEFINE ANALYZER {self._name}{tok}{filters}"


class DefineScope(Statement):
    """
    Represents DEFINE SCOPE statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/scope

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    """

    def __init__(self, connection: Connection, name: str, duration: str, signup: Union[str, Statement],
                 signin: Union[str, Statement]):
        super().__init__(connection)
        self._name = name
        self._duration = duration
        self._signup = signup if not isinstance(signup, Statement) else f"({signup._clean_str()})"
        self._signin = signin if not isinstance(signin, Statement) else f"({signin._clean_str()})"

    def validate(self) -> List[str]:
        if ' ' in self._duration:
            return ["Wrong duration format, should be like 24h"]
        return [OK]

    def _clean_str(self):
        return f"DEFINE SCOPE {self._name} SESSION {self._duration} \nSIGNUP {self._signup} \nSIGNIN {self._signin}"


class DefineIndex(Statement):
    """
    Represents DEFINE INDEX statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/indexes

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    """

    def __init__(self, connection: Connection, name: str, table_name: str):
        super().__init__(connection)
        self._name = name
        self._table_name = table_name
        self._fields = None
        self._uni = False
        self._analyzer = None

    def fields(self, fields: str) -> "DefineIndex":
        """
        Represents fields for index, FIELDS statement

        :param fields: fields to index
        """
        self._fields = f"FIELDS {fields}"
        return self

    def columns(self, columns: str) -> "DefineIndex":
        """
        Represents fields for index, COLUMNS statement

        :param columns: fields to index
        """
        self._fields = f"COLUMNS {columns}"
        return self

    def unique(self) -> "DefineIndex":
        """
        Represents UNIQUE statement
        """
        self._analyzer = None
        self._uni = True
        return self

    def search_analyzer(self, name: str, use: str = "BM25", highlights: bool = True) -> "DefineIndex":
        """
        Represents SEARCH ANALYZER statement
        """
        self._uni = False
        hl = "" if not highlights else " HIGHLIGHTS"
        self._analyzer = f" SEARCH ANALYZER {name} {use}{hl}"
        return self

    def validate(self) -> List[str]:
        if not self._fields:
            return ["You need to specify FIELDS or COLUMNS"]
        return [OK]

    def _clean_str(self):
        add = ""
        if self._uni:
            add = " UNIQUE"
        elif self._analyzer:
            add = self._analyzer
        return f"DEFINE INDEX {self._name} ON TABLE {self._table_name} {self._fields}{add}"


class DefineToken(Statement):
    """
    Represents DEFINE TOKEN statement

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/token

    Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    """

    def __init__(self, connection: Connection, name: str, token_type: str, value: str):
        super().__init__(connection)
        self._name = name
        self._type = token_type
        self._value = value

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        return f'DEFINE TOKEN {self._name} ON DATABASE \nTYPE {self._type} \nVALUE "{self._value}"'



from typing import List, Optional

from surrealist import Connection
from surrealist.ql.statements.statement import Statement
from surrealist.ql.statements.update_statements import UpdateUseMethods
from surrealist.utils import OK


class Update(Statement, UpdateUseMethods):
    """
    Represents UPDATE statement, it should be able to use any statements from documentation

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/update

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_update_examples.py
    """

    def __init__(self, connection: Connection, table_name: str, record_id: Optional[str] = None):
        super().__init__(connection)
        self._table_name = table_name
        self._only = False
        self._record_id = record_id

    def validate(self) -> List[str]:
        return [OK]

    def only(self) -> "Update":
        """
        Include ONLY statement for the query
        """
        self._only = True
        return self

    def _clean_str(self):
        what = "" if not self._only else " ONLY"
        table = self._table_name if not self._record_id else f"{self._table_name}:{self._record_id}"
        return f"UPDATE{what} {table}"



from surrealist.result import SurrealResult
from surrealist.ql.statements.statement import FinishedStatement, Statement


class Fetch(FinishedStatement):
    def __init__(self, statement: Statement, *args: str):
        super().__init__(statement)
        self._args = args

    def _clean_str(self):
        if not self._args:
            return self._statement._clean_str()
        what = ", ".join(self._args)
        return f"{self._statement._clean_str()} FETCH {what}"

    def run(self) -> SurrealResult:
        return self._statement._drill(self.to_str())


class LiveUseFetch:
    def fetch(self, *args: str) -> Fetch:
        return Fetch(self, *args)


class Where(FinishedStatement, LiveUseFetch):
    def __init__(self, statement: Statement, predicate: str):
        super().__init__(statement)
        self._predicate = predicate

    def run(self) -> SurrealResult:
        return self._statement._drill(self.to_str())

    def _drill(self, query) -> SurrealResult:
        return self._statement._drill(query)

    def _clean_str(self):
        return f"{self._statement._clean_str()} WHERE {self._predicate}"


class LiveUseWhere(LiveUseFetch):
    def where(self, predicate: str) -> Where:
        return Where(self, predicate)



from abc import ABC, abstractmethod
from typing import List, Iterator

from surrealist.connections import Connection
from surrealist.result import SurrealResult
from surrealist.utils import OK


class Statement(ABC):
    """
    Parent for all statements(QL statements)
    """

    def __init__(self, connection: Connection):
        self._connection = connection

    def _validate(self) -> str:
        return OK

    @abstractmethod
    def validate(self) -> List[str]:
        """
        Should check all parts of the query ant returns list of OK or errors
        :return: list of strings
        """

    def is_valid(self) -> bool:
        """
        Checks is whole query is valid
        :return: True, f query is valid, False otherwise
        """
        return all(mess == OK for mess in self.validate())

    @abstractmethod
    def _clean_str(self) -> str:
        """
        Returns query without ";" to work with
        :return: query text
        """

    def to_str(self):
        """
        Returns the whole query with ";" at the end
        :return: full query text
        """
        return f"{self._clean_str()};"

    def _drill(self, query) -> SurrealResult:
        """
        This method for live queries only
        :param query: full query text
        :return: result of the query
        """

    def run(self) -> SurrealResult:
        """
        Runs the whole query and returns result from SurrealDB
        :return: result of the request
        """
        return self._connection.query(self.to_str())

    def __str__(self):
        return self.to_str()


class FinishedStatement(Statement):
    """
    Represents specific statement
    """

    def __init__(self, statement: Statement):
        self._connection = statement._connection
        super().__init__(self._connection)
        self._statement = statement

    def validate(self) -> List[str]:
        result = self._statement.validate()
        result.append(self._validate())
        return [e for e in result if e != OK]

    def _clean_str(self):
        return self._statement._clean_str()


class IterableStatement(FinishedStatement):
    """
    Represents a statement which can use iterator to get results in an efficient, lazy way.
    Under the hood transform query to SELECT * FROM (initial_query) LIMIT {limit} START AT {current};
    """

    def iter(self, limit: int = 100) -> Iterator:
        """
        Creates and returns a generator object to iterate on big query results

        In documentation: https://github.com/kotolex/surrealist/tree/master?tab=readme-ov-file#iteration-on-select

        Example: https://github.com/kotolex/surrealist/tree/master/examples/surreal_ql/iterator.py

        :param limit: number of records in each iteration, it cannot be smaller than one
        :return: generator to use in for-statements or with the next method
        :raise ValueError: if limit less than one
        """
        if limit < 1:
            raise ValueError("The limit cannot be smaller than 1")
        current = 0
        while True:
            query = f"SELECT * FROM ({self._clean_str()}) LIMIT {limit} START AT {current};"
            res = self._connection.query(query)
            yield res
            if res.count() < limit:
                break
            current += limit



from surrealist.ql.statements.statement import FinishedStatement, Statement


class OnDuplicate(FinishedStatement):
    def __init__(self, statement: Statement, action: str):
        super().__init__(statement)
        self._action = action

    def _clean_str(self):
        return f"{self._statement._clean_str()} ON DUPLICATE KEY UPDATE {self._action}"


class InsertUseDuplicate:
    def on_duplicate(self, action: str) -> OnDuplicate:
        return OnDuplicate(self, action)



from surrealist.ql.statements.statement import FinishedStatement, Statement
from surrealist.utils import OK


class Parallel(FinishedStatement):

    def _clean_str(self):
        return f"{self._statement._clean_str()} PARALLEL"


class CanUseParallel:
    def parallel(self) -> Parallel:
        return Parallel(self)


class Timeout(FinishedStatement, CanUseParallel):
    def __init__(self, statement: Statement, duration: str):
        super().__init__(statement)
        self._duration = duration

    def _validate(self):
        if not any(self._duration.endswith(letter) for letter in ('ms', 's', 'm')):
            return f"Wrong duration {self._duration}, allowed postfix are (ms, s, m)"
        if ' ' in self._duration:
            return "Wrong duration format, should be like 5s"
        return OK

    def _clean_str(self):
        return f"{self._statement._clean_str()} TIMEOUT {self._duration}"


class CanUseTimeout(CanUseParallel):
    def timeout(self, duration: str) -> Timeout:
        return Timeout(self, duration)


class ReturnNone(FinishedStatement, CanUseTimeout):
    def _clean_str(self):
        return f"{self._statement._clean_str()} RETURN NONE"


class ReturnBefore(FinishedStatement, CanUseTimeout):

    def _clean_str(self):
        return f"{self._statement._clean_str()} RETURN BEFORE"


class ReturnAfter(FinishedStatement, CanUseTimeout):

    def _clean_str(self):
        return f"{self._statement._clean_str()} RETURN AFTER"


class ReturnDiff(FinishedStatement, CanUseTimeout):

    def _clean_str(self):
        return f"{self._statement._clean_str()} RETURN DIFF"


class Return(FinishedStatement, CanUseTimeout):
    def __init__(self, statement: Statement, *args: str):
        super().__init__(statement)
        self._args = args

    def _clean_str(self):
        if not self._args:
            return self._statement._clean_str()
        args = ", ".join(self._args)
        return f"{self._statement._clean_str()} RETURN {args}"


class CanUseReturn(CanUseTimeout):
    def return_diff(self) -> ReturnDiff:
        return ReturnDiff(self)

    def return_none(self) -> ReturnNone:
        return ReturnNone(self)

    def return_before(self) -> ReturnBefore:
        return ReturnBefore(self)

    def return_after(self) -> ReturnAfter:
        return ReturnAfter(self)

    def returns(self, *args: str) -> Return:
        return Return(self, *args)


class Where(FinishedStatement, CanUseReturn):
    def __init__(self, statement: Statement, predicate: str):
        super().__init__(statement)
        self._predicate = predicate

    def _clean_str(self):
        return f"{self._statement._clean_str()} WHERE {self._predicate}"


class CanUseWhere(CanUseReturn):
    def where(self, predicate: str) -> Where:
        return Where(self, predicate)



from typing import Optional, Tuple, Union, List

from surrealist import Connection
from surrealist.utils import OK
from .select_statements import SelectUseIndex
from .statement import Statement, IterableStatement


class Select(IterableStatement, SelectUseIndex):
    """
    Represents SELECT statement, it should be able to use any statements from documentation.
    It can use iterator to traverse results

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/select

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_select_examples.py

    Iterator example: https://github.com/kotolex/surrealist/tree/master/examples/surreal_ql/iterator.py
    """

    def __init__(self, connection: Connection, table_name: Union[str, Statement], *args,
                 alias: Optional[Tuple[str, Union[str, Statement]]] = None,
                 value: Optional[str] = None):
        self._connection = connection
        super().__init__(self)
        self._table_name = table_name if isinstance(table_name, str) else f"({table_name._clean_str()})"
        self._args = args
        self._what = ""
        self._from = self._table_name
        self._only = False
        self._value = value
        self._as = alias
        self._update_args()
        self._omit = None

    def _update_args(self):
        if self._args:
            self._what = ", ".join(self._args)
        if self._as:
            what = "" if not self._what else f"{self._what}, "
            value = self._as[1] if not isinstance(self._as[1], Statement) else f"({self._as[1]._clean_str()})"
            self._what = f"{what}{value} AS {self._as[0]}"
        if self._value:
            self._what = f"VALUE {self._value}"

    def by_id(self, record_id: str) -> "Select":
        """
        Select table_name:record_id

        :param record_id: id of the record to select
        """
        self._from = f"{self._table_name}:{record_id}"
        return self

    def only(self) -> "Select":
        """
        Include ONLY statement for the query
        """
        self._only = True
        return self

    def omit(self, *fields: Tuple[str]) -> "Select":
        """
        Include OMIT statement for the query
        :param fields: fields to omit
        """
        self._omit = fields
        return self

    def validate(self) -> List[str]:
        if self._only and ":" not in self._from:
            return ["Expected a single result output when using the ONLY keyword"]
        return [OK]

    def _clean_str(self):
        only = "ONLY " if self._only else ""
        what = "*" if not self._what else self._what
        omit = ""
        if self._omit:
            omitted = ", ".join(self._omit)
            omit = f" OMIT {omitted}"
        return f"SELECT {what}{omit} FROM {only}{self._from}"

    def __repr__(self):
        return f"Select({self._table_name=}, {self._args=}, {self._as=}, {self._value=})"



import re
import uuid
from typing import Union

ENCODING = "UTF-8"
OK = "OK"
ERR = "ERR"
DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
DATE_FORMAT_NS = "%Y-%m-%dT%H:%M:%S.%fZ"
HTTP_OK = 200  # status code for success
DEFAULT_TIMEOUT = 5  # timeout in seconds for basic operations
DATA_LENGTH_FOR_LOGS = 300  # size of data in logs, data will be cropped if bigger than that


def _set_length(length: int):
    global DATA_LENGTH_FOR_LOGS
    DATA_LENGTH_FOR_LOGS = length


def get_uuid() -> str:
    """
    Helper to generate uuid for records or testing purposes

    :return: uuid string representation
    """
    return str(uuid.uuid4())


def mask_pass(text: str) -> str:
    """
    Mask the passwords in logs. Replace all 'pass':'any_pass' in logs to 'pass':'******'. Works both with ' and "
    :param text: text before putting it to log
    :return: text without visible passwords
    """
    return re.sub(r"(?ms)(?<=['\"]pass['\"]: ['\"]).*?(?=['\"])", '******', text)


def crop_data(data: Union[str, bytes], is_str: bool = True) -> Union[str, bytes]:
    """
    Crop the data to maximum size, no actions if data is smaller, work with str and bytes in logs

    :param data: str or bytes of data to put it in the logs
    :param is_str: a flag to use correct operation
    :return: data of a same type, but cropped if it is bigger than DATA_LENGTH_FOR_LOGS
    """
    if len(data) > DATA_LENGTH_FOR_LOGS:
        return f"{data[:DATA_LENGTH_FOR_LOGS]}..." if is_str else data[:DATA_LENGTH_FOR_LOGS] + b'...'
    return data



from typing import List

from surrealist.connections import Connection
from surrealist.ql.statements.create_statements import CreateUseSetContent
from surrealist.ql.statements.statement import Statement
from surrealist.utils import OK


class Relate(Statement, CreateUseSetContent):
    """
    Represents RELATE statement, it should be able to use any statements from documentation

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/relate

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_relate_examples.py

    """

    def __init__(self, connection: Connection, value: str):
        """
        Init for Create
        :param connection: connection relied on
        :param value: relate representation, see examples
        """
        super().__init__(connection)
        self._only = False
        self._value = value

    def only(self) -> "Relate":
        """
        Include ONLY statement for the query
        """
        self._only = True
        return self

    def validate(self) -> List[str]:
        if "->" not in self._value and "<-" not in self._value:
            return ["Wrong format for RELATE representation"]
        return [OK]

    def _clean_str(self):
        only = "" if not self._only else " ONLY"
        return f"RELATE{only} {self._value}"



import logging
from typing import Optional, Tuple, List, Dict, Union, Any, Callable

from surrealist.ql.statements import Select, Remove, Live
from surrealist.ql.statements.define import (DefineEvent, DefineUser, DefineParam, DefineAnalyzer, DefineScope,
                                             DefineIndex, DefineToken)
from surrealist.ql.statements.relate import Relate
from surrealist.ql.statements.returns import Return
from surrealist.ql.statements.statement import Statement
from surrealist.ql.statements.transaction import Transaction
from surrealist.ql.table import Table
from surrealist.result import SurrealResult
from surrealist.surreal import Surreal
from surrealist.utils import DEFAULT_TIMEOUT

logger = logging.getLogger("databaseQL")


class Database:
    """
    Represents connected database(in some namespace) to operate on.
    It has features of the database level, including switch to table level. You can use DEFINE/REMOVE, query or
    transactions here, but for simple CRUD you need to switch to table level

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    """

    def __init__(self, url: str, namespace: str, database: str, credentials: Optional[Tuple[str, str]],
                 use_http: bool = False, timeout: int = DEFAULT_TIMEOUT, log_level: str = "ERROR"):
        self._namespace = namespace
        self._database = database
        self._connection = Surreal(url, namespace, database, credentials, use_http, timeout, log_level).connect()
        self._connected = True
        logger.info("DatabaseQL is up")

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        self.close()

    def is_connected(self) -> bool:
        """
        Return if a database is connected
        :return: True if connection is active, False otherwise
        """
        return self._connected

    @property
    def namespace(self) -> str:
        """
        Return name if the namespace
        :return: namespace
        """
        return self._namespace

    @property
    def name(self) -> str:
        """
        Return name of the database
        :return: database name
        """
        return self._database

    def close(self):
        """
        Closes the connection. You cannot and should not use a database object after that
        """
        logger.info("DatabaseQL is down")
        self._connection.close()
        self._connected = False

    def tables(self) -> List[str]:
        """
        Return list of the table names at a current database
        :return: string list with the names
        """
        logger.info("Get tables for db %s", self._database)
        return self._connection.db_tables().result

    def info(self) -> Dict:
        """
        Return full info about a database, actually call for "INFO FOR DB;" via query
        :return: a result of the response
        """
        logger.info("Get info for db %s", self._database)
        return self._connection.db_info().result

    def raw_query(self, query: str) -> SurrealResult:
        """
        Execute a raw QL query on a database, you should use it in cases when your query is large or complicated, or it
        is just more appropriate for you.

        :param query: text of the query, it will be sent to SurrealDB as is
        :return: a result of the response
        """
        logger.info("Query for db %s", self._database)
        return self._connection.query(query)

    def returns(self, query: str) -> Return:
        """
        Return result of the query

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/return

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param query: string query to execute
        :return: Return object
        """
        return Return(self._connection, query)

    def __getattr__(self, item) -> Table:
        """
        We actually need this to use dot as the switch to table level as "db.person" instead of "db.table("person")"
        :param item: name of the table to work with
        :return: a table object
        """
        return Table(item, self._connection)

    def table(self, name) -> Table:
        """
        Switch to the Table level(an object) after that you can and should use table operations (CRUD)

        :param name: name of the table to work with
        :return: a table object
        """
        return Table(name, self._connection)

    def transaction(self, statements: List[Statement]) -> Transaction:
        """
        Create a transaction object to generate a query or run

        Refer to: https://docs.surrealdb.com/docs/surrealql/transactions

        :param statements: list of appropriate statements (select, create, delete. etc.)
        :return: Transaction object
        """
        return Transaction(self._connection, statements)

    def select_from(self, select: Select, *args, alias: Optional[Tuple[str, Union[str, Statement]]] = None,
                    value: Optional[str] = None) -> Select:
        """
        Allow selecting from sub-query (select)

        :param select: Select object
        :param args: which fields to select, if no fields or "*" - selects all
        :param alias: pairs of names and values
        :param value: on exists add VALUE statement
        :return: Select object
        """
        return Select(self._connection, select, *args, alias=alias, value=value)

    def define_event(self, name: str, table_name: str, then: Union[str, Statement]) -> DefineEvent:
        """
        Allow defining event on table

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/event

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the event
        :param table_name: name of the table
        :param then: action to perform
        :return: DefineEvent object
        """
        return DefineEvent(self._connection, name, table_name, then)

    def remove_event(self, name: str, table_name: str) -> Remove:
        """
        Remove an event linked to table

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/remove

        :param name: name of the event
        :param table_name: name of the table
        :return: Remove object
        """
        return Remove(self._connection, table_name=table_name, type_="EVENT", name=name)

    def define_user(self, user_name: str, password: str) -> DefineUser:
        """
        Allow defining user for a current database

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/user

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param user_name: name for the new user
        :param password: password for user
        :return: DefineUser object
        """
        return DefineUser(self._connection, user_name=user_name, password=password)

    def remove_user(self, user_name: str) -> Remove:
        """
        Remove user of the database

        :param user_name: name of the user
        :return: Remove object
        """
        return Remove(self._connection, "", type_="USER", name=user_name)

    def define_param(self, name: str, value: Any) -> DefineParam:
        """
        Represents DEFINE PARAM statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/param

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name of the parameter
        :param value: value for the parameter
        :return: DefineParam object
        """
        return DefineParam(self._connection, name, value)

    def remove_param(self, name: str) -> Remove:
        """
        Remove parameter of the database

        :param name: name of the parameter
        :return: Remove object
        """
        return Remove(self._connection, "", type_="PARAM", name=name)

    def define_analyzer(self, name: str) -> DefineAnalyzer:
        """
        Represents DEFINE ANALYZER statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/analyzer

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the analyzer
        :return: DefineAnalyzer object
        """
        return DefineAnalyzer(self._connection, name)

    def remove_analyzer(self, name: str) -> Remove:
        """
        Remove the analyzer

        :param name: name of the analzer
        :return: Remove object
        """
        return Remove(self._connection, '', type_="ANALYZER", name=name)

    def define_scope(self, name: str, duration: str, signup: Union[str, Statement],
                     signin: Union[str, Statement]) -> DefineScope:
        """
        Represents DEFINE SCOPE statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/scope

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the new scope
        :param duration: session duration, like 24h
        :param signup: Create statement with string or Statement representation
        :param signin: Select statement with string or Statement representation
        :return: DefineScope object
        """
        return DefineScope(self._connection, name, duration, signup, signin)

    def remove_scope(self, name: str) -> Remove:
        """
        Remove the scope

        :param name: name of the scope
        :return: Remove object
        """
        return Remove(self._connection, "", type_="SCOPE", name=name)

    def define_index(self, name: str, table_name: str) -> DefineIndex:
        """
        Represents DEFINE INDEX statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/indexes

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the index
        :param table_name: name of table to work with
        :return: DefineINdex object
        """
        return DefineIndex(self._connection, name, table_name)

    def remove_index(self, name: str, table_name: str) -> Remove:
        """
        Remove index by name on the table

        :param name: name og the index
        :param table_name: name of the table
        :return: Remove object
        """
        return Remove(self._connection, name=name, table_name=table_name, type_="INDEX")

    def define_token(self, name: str, token_type: str, value: str) -> DefineToken:
        """
        Represents DEFINE TOKEN statement

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/define/token

        Example: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py

        :param name: name for the token
        :param token_type: type of the token, for example, RS256
        :param value: value of the token
        :return: DefineINdex object
        """
        return DefineToken(self._connection, name, token_type, value)

    def remove_token(self, name: str) -> Remove:
        """
        Remove token by name for the database

        :param name: name of the token
        :return: Remove object
        """
        return Remove(self._connection, "", type_="TOKEN", name=name)

    def relate(self, value: str) -> Relate:
        """
        Represents RELATE statement, it should be able to use any statements from documentation

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/relate

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_relate_examples.py

        :param value: relate representation, see examples
        :return: Relate object
        """
        return Relate(self._connection, value=value)

    def live_query(self, table_name: str, callback: Callable[[Dict], None], use_diff: bool = False) -> Live:
        """
        Represents LIVE statement for a live query

        Example:
        db.live_query("person", func).alias("first_name", "NAME").where("age > 22").run()

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/live-select

        Refer to: https://github.com/kotolex/surrealist?tab=readme-ov-file#live-query

        Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_live_examples.py

        :param table_name: name od the table to live select
        :param callback: function to call on live query event, signature is `def callback(arg:Dict) -> None`
        :param use_diff: return result in DIFF format
        :return: Live object
        """
        return Live(self._connection, table_name, callback, use_diff)

    def kill_query(self, live_id: str) -> SurrealResult:
        """
        Represents a KILL statement, for killing a live query by id

        Refer to: https://docs.surrealdb.com/docs/surrealql/statements/kill

        :param live_id: id of the query
        :return: result
        """
        return self._connection.kill(live_id)

    def __repr__(self):
        return f"Database(namespace={self._namespace}, name={self._database}, connected={self.is_connected()})"



from typing import List, Optional

from surrealist import Connection
from surrealist.ql.statements.statement import Statement
from surrealist.utils import OK


class Remove(Statement):
    """
    Represents REMOVE statement, it should be able to use any statements from documentation

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/remove

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/database.py
    """
    _variants = ("TABLE", "EVENT", "FIELD", "INDEX", "PARAM", "USER", "ANALYZER", "SCOPE", "TOKEN")

    def __init__(self, connection: Connection, table_name: str, type_: str = "TABLE", name: Optional[str] = None):
        if type_ not in Remove._variants:
            raise ValueError(f"Type should be one of {Remove._variants}")
        super().__init__(connection)
        self._table_name = table_name
        self._only = False
        self._type = type_
        self._name = name
        if type_ == "TABLE":
            self._name = table_name

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        if self._type in ("TABLE", "PARAM", "USER", "ANALYZER", "SCOPE", "TOKEN"):
            what = f"{self._type} {self._name}"
            if self._type in ("USER", "TOKEN"):
                what = f"{what} ON DATABASE"
            if self._type == "PARAM":
                what = f"{self._type} ${self._name}"
            if self._type == "TABLE":
                what = f"{what} IF EXISTS"
        else:
            what = f"{self._type} {self._name} ON TABLE {self._table_name}"
        return f"REMOVE {what}"



from .create import Create
from .define import DefineEvent, DefineParam, DefineUser
from .delete import Delete
from .insert import Insert
from .live import Live
from .remove import Remove
from .select import Select
from .show import Show
from .transaction import Transaction
from .update import Update

__all__ = ("Create", "Delete", "Insert", "Live", "Remove", "Select", "Show", "Update", "Transaction", "DefineParam",
           "DefineUser", "DefineEvent")



from typing import List, Optional

from surrealist.connections import Connection
from surrealist.ql.statements.create_statements import CreateUseSetContent
from surrealist.ql.statements.statement import Statement
from surrealist.utils import OK


class Create(Statement, CreateUseSetContent):
    """
    Represents CREATE statement, it should be able to use any statements from documentation

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/create

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_create_examples.py

    """

    def __init__(self, connection: Connection, table_name: str, record_id: Optional[str] = None):
        """
        Init for Create
        :param connection: connection relied on
        :param table_name: name of the table to operate
        :param record_id: optional, if exist-transform to 'table_name:record_id'
        """
        super().__init__(connection)
        self._table_name = table_name
        self._only = False
        self._record_id = record_id

    def only(self) -> "Create":
        """
        Include ONLY statement for the query
        """
        self._only = True
        return self

    def validate(self) -> List[str]:
        return [OK]

    def _clean_str(self):
        only = "" if not self._only else " ONLY"
        name = self._table_name if not self._record_id else f"{self._table_name}:{self._record_id}"
        return f"CREATE{only} {name}"



from typing import List, Callable

from surrealist.connections import Connection
from surrealist.ql.statements.live_statements import LiveUseWhere
from surrealist.ql.statements.statement import Statement
from surrealist.result import SurrealResult
from surrealist.utils import OK


class Live(Statement, LiveUseWhere):
    """
    Represents LIVE SELECT statement, it should be able to use any statements from documentation

    Refer to: https://docs.surrealdb.com/docs/surrealql/statements/live-select

    Examples: https://github.com/kotolex/surrealist/blob/master/examples/surreal_ql/ql_live_examples.py
    """

    def __init__(self, connection: Connection, table_name: str, callback: Callable, use_diff: bool = False):
        super().__init__(connection)
        self._table_name = table_name
        self._alias = None
        self._diff = use_diff
        self._callback = callback

    def alias(self, value_name: str, alias: str) -> "Live":
        """

        :param value_name:
        :param alias:
        :return:
        """
        self._alias = (value_name, alias)
        self._diff = False
        return self

    def validate(self) -> List[str]:
        return [OK]

    def run(self) -> SurrealResult:
        return self._drill(self.to_str())

    def _drill(self, query):
        return self._connection.custom_live(query, self._callback)

    def _clean_str(self):
        if self._diff:
            what = "DIFF"
        elif self._alias:
            what = f"{self._alias[0]} AS {self._alias[1]}"
        else:
            what = "*"
        return f"LIVE SELECT {what} FROM {self._table_name}"
